# -*- coding: utf-8 -*-
# PYTHON_ARGCOMPLETE_OK

# EeSSH -- SSH client for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eessh
#
# This file is part of EeSSH.
#
# EeSSH is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EeSSH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
import argparse, argcomplete
import os.path
import sys

from eessh.api import api
from eessh.cache import xdgcache
from eessh.client import EeSSHClient
from eessh.config import config

log = logging.getLogger('eessh.cli')


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    # Default configuration file
    default_config_filepath = os.path.join(os.path.expanduser('~'), '.config', 'eemyle', 'eesshrc') # XXX hard-coded path (and repeated elsewhere)

    parser = argparse.ArgumentParser()

    parser.add_argument(
        '-a',
        '--action',
        default='shell',
        choices=["shell", "eeshell", "show-tickets", "show-passwords", "show-comments"],
        help='Action to launch (default: shell)'
    )

    parser.add_argument(
        '-A',
        '--agent-forwarding',
        action='store_true',
        help='Enable agent forwarding'
    )

    parser.add_argument(
        '-c',
        '--config',
        default=default_config_filepath,
        help='Configuration file to use (default is {0})'.format(default_config_filepath)
    )

    parser.add_argument(
        '-d',
        '--debug',
        action='store_true',
        help='Show debug messages'
    )

    parser.add_argument(
        '-l',
        '--list',
        action='store_true',
        help='List customer\'s servers'
    )

    parser.add_argument(
        '-n',
        '--no-firsthop',
        action='store_true',
        help='Skip first hop'
    )

    parser.add_argument(
        '-b',
        '--broadcast',
        action='store_true',
        help='Enable broadcast mode (default is disabled)'
    )

    parser.add_argument(
        '-r',
        '--resolvable',
        action='store_true',
        help='Argument is a resolvable hostname'
    )

    parser.add_argument(
        '-p',
        '--port',
        type=int,
        help='Port for resovable hostname'
    )

    parser.add_argument(
        '-C',
        '--compress',
        default=False,
        action='store_true',
        help='Enable compression'
    )

    parser.add_argument(
        '-u',
        '--username',
        help='Specifies the user to log in as on the remote machine.'
    )

    parser.add_argument(
        '-X',
        '--x11-forwarding',
        default=False,
        action='store_true',
        help='Enable X11 forwarding'
    )

    mainarg = parser.add_argument(
        'hostname',
    )

    def _autocomplete_hostnames(prefix, parsed_args, **kwargs):
        if not xdgcache.get("_API.get_servernames_"):
            from argcomplete import warn
            warn("Initializing autocompletion cache (using API settings from {}). Please wait...".format(default_config_filepath))
            try:
                config.load(default_config_filepath)
                api.initialize()
                api.get_servernames()
                warn("Cache loaded.")
            except Exception as e:
                warn(e)
        return (h for h in api.get_servernames() if h.startswith(prefix))
    mainarg.completer = _autocomplete_hostnames

    argcomplete.autocomplete(parser)
    options = parser.parse_args(argv)

    if options.config:
        # Getting configuration file from command
        config_filepath = os.path.abspath(options.config)

    # Ensure personal config file is present
    if not os.path.isfile(config_filepath):
        log.warning("Configuration file is missing. {}".format(config_filepath))
        cfg_parentdir = os.path.dirname(config_filepath)
        if not os.path.isdir(cfg_parentdir):
            os.makedirs(cfg_parentdir)
        with open(config_filepath, 'w') as fout:
            fout.write('''[api]
url = https://ws.easter-eggs.fr/ws/eemyle.php
username = eeadmin
legacy = true

[firsthop]
hostname = vpn-clients.easter-eggs.fr
username = eeadmin

[connection]
agent_forwarding = true''')


    # Load configuration file
    if not config.load(config_filepath):
        parser.error('Failed to load configuration from file %s' % config_filepath)

    if options.debug:
        logging.getLogger().setLevel(logging.DEBUG)

    # Parse argument
    hostname = options.hostname

    if options.resolvable:
        if '@' in hostname:
            username, hostname = hostname.split('@')
        else:
            username = options.username

    elif '.' in hostname:
        hostname, customer_uid = hostname.rsplit('.', 1)
        if not hostname:
            parser.error('Missing hostname')

        if not customer_uid:
            parser.error('Missing customer Id')

    elif options.list:
        customer_uid = hostname

    else:
        parser.error('Failed to exract customer_uid from hostname')

    if not api.initialize():
        parser.error('Failed to initialize API connector')

    if options.list:
        # Get servers list
        servers = api.get_servers(customer_uid)
        if servers is None:
            parser.error('Failed to get servers list for customer {0}'.format(customer_uid))

        for server in servers:
            print(' * {hostname:30s} {os_name:40s} {description:50s}'.format(**server))

        return True

    # Use cases where we don't need to launch a client
    if options.action in ("show-tickets", "show-passwords", "show-comments"):
        import json
        info = api.get_data(customer_uid, hostname)
        if options.action == "show-passwords":
            print(json.dumps(info.get("users"), indent=True))
        elif options.action == "show-tickets":
            for ticket in api.get_tickets(customer_uid, hostname):
                print("""#{number} {subject}""".format(**ticket))
            # print(json.dumps(res, indent=True))
        elif options.action == "show-comments":
            print(info.get("comments"))
        else:
            print("Unknow option {}".format(options.action))
            print(json.dumps(info, indent=True))
        return True

    # Flush screen
    sys.stdout.write('\x1b[2J\x1b[H\n\n')
    sys.stdout.flush()

    if options.resolvable:
        data = dict(
            customer_uid=None,
            hostname=hostname,
            hops=[
                dict(
                    host=hostname,
                    port=options.port or 22,
                    hostname=hostname,
                    username=username,
                )
            ],
            ipmi=None,
        )
    else:
        # Get HOPS
        sys.stdout.write('         Getting connection parameters from GEemyle server...')
        sys.stdout.flush()

        data = api.get_data(customer_uid, hostname)
        if data is None:
            sys.stderr.write('\n\r[\x1b[31mFAILED\x1b[39m] Failed to get connection parameters from GEemyle server.')
            return False

        sys.stdout.write('\n\r[  \x1b[32mOK\x1b[39m  ] Got connection parameters, needs to connect to {0} hop(s)...\n\r'.format(len(data['hops'])))
        sys.stdout.flush()

    # Launch client
    client = EeSSHClient(
        data,
        agent_forwarding=options.agent_forwarding,
        broadcast=options.broadcast,
        compress=options.compress,
        no_firsthop=options.no_firsthop,
        username=options.username,
        x11_forwarding=options.x11_forwarding
    )

    client.run(options.action)
