# -*- coding: utf-8 -*-

# EeSSH -- SSH client for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eessh
#
# This file is part of EeSSH.
#
# EeSSH is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EeSSH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" API connector """

import logging

from eessh.config import config
from eessh.legacy import LegacyConnector
from eessh.cache import xdgcache as region

log = logging.getLogger('eessh.api')


# Register default options
DEFAULTS = {
    'debug': (bool, False),
    'fingerprint': (str, None),
    'legacy': (bool, False),
    'password': (str, None),
    'url': (str, None),
    'username': (str, None),
    'verify': (bool, True),
}

config.set_defaults('api', DEFAULTS)


class API:
    def __init__(self):
        self.__connector = None

    @property
    def ready(self):
        return self.__connector is not None

    @region.cache_on_arguments()
    def get_customers(self, *args, **kwargs):
        assert self.__connector is not None, 'API not initialized'

        with self.__connector as connector:
            return connector.get_customers(*args, **kwargs)

    @region.cache_on_arguments()
    def get_data(self, *args, **kwargs):
        assert self.__connector is not None, 'API not initialized'

        with self.__connector as connector:
            return connector.get_data(*args, **kwargs)

    def get_ipmi(self, *args, **kwargs):
        assert self.__connector is not None, 'API not initialized'

        with self.__connector as connector:
            return connector.get_ipmi(*args, **kwargs)

    @region.cache_on_arguments()
    def get_server(self, *args, **kwargs):
        assert self.__connector is not None, 'API not initialized'

        with self.__connector as connector:
            return connector.get_server(*args, **kwargs)

    @region.cache_on_arguments()
    def get_servers(self, *args, **kwargs):
        assert self.__connector is not None, 'API not initialized'

        with self.__connector as connector:
            return connector.get_servers(*args, **kwargs)

    def get_tickets(self, *args, **kwargs):
        assert self.__connector is not None, 'API not initialized'

        with self.__connector as connector:
            return connector.get_tickets(*args, **kwargs)

    @region.cache_on_arguments()
    def get_servernames(self):
        """
        Return servernames as a list instead of a generator.
        Only useful for caching + autocompletion.
        Prefer function `sefl.iter_all_servernames()`.
        """
        return [h for h in self.iter_all_servernames()]

    def iter_all_servernames(self):
        for customer in self.get_customers():
            eeid = customer.get('customer_uid')
            for server in self.get_servers(customer_uid=eeid):
                servername = server.get('hostname')
                fqname = "{0}-{1}".format(eeid, servername)
                yield fqname

    def invalidate_cache(self):
        region.invalidate()

    def initialize(self):
        """ Initialize connector """

        if config.get('api', 'legacy') is True:
            class_ = LegacyConnector
        else:
            log.error('Only legacy mode is supported for now')
            return False

        url = config.get('api', 'url')
        if url is None:
            log.error('Missing API url, please check configuration')
            return False

        options = dict(
            (option, config.get('api', option))
            for option in ('debug', 'password', 'username', 'verify')
        )

        self.__connector = class_(url, **options)

        return True


# Initialize API object
api = API()
