# -*- coding: utf-8 -*-

# EeSSH -- SSH client for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eessh
#
# This file is part of EeSSH.
#
# EeSSH is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EeSSH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" SSH Client """

import errno
import fcntl
import logging
import os
import re
import select
import socket
import stat
import struct
import sys
import tempfile
import termios
import time
import tty

from paramiko import AutoAddPolicy
from paramiko import SSHClient
from paramiko.agent import AgentRequestHandler
from paramiko.util import ClosingContextManager

from eessh.config import config
from eessh.forwarding import CForwarder
from eessh.forwarding import PForwarder
from eessh.forwarding import RForwarder
from eessh.forwarding import XForwarder
from eessh.utils import attr_data
from eessh.utils import KB
from eessh.utils import MasterFifo
from eessh.utils import size_hr
from eessh.utils import SlaveFifo
from eessh.utils import time_hr


log = logging.getLogger('eessh.client')

tempdir = tempfile.gettempdir()


# Register default options
DEFAULTS = {
    'enabled': (bool, False),
    'host': (str, None),
    'port': (int, 22),
    'hostname': (str, None),
    'username': (str, None),
    'password': (str, None),
}

config.set_defaults('firsthop', DEFAULTS)


class EeSSHClient:

    CONNECT_TIMEOUT = 10

    CTRL_FIFO = os.path.join(tempdir, 'eessh-ctrl')
    DATA_FIFO = os.path.join(tempdir, 'eessh-data')

    READ_SIZE = 1024

    def __init__(self, data, agent_forwarding=False, broadcast=False, compress=False, embedded=False, no_firsthop=False, username=None, x11_forwarding=False):

        self.__afwd = None
        self.__data = data
        self.__mode = None
        self.__loop = True
        self.__cfwd = None
        self.__pfwds = []
        self.__pfwds_index = 0
        self.__rfwds = []
        self.__rfwds_index = 0
        self.__sftp = None
        self.__shell = False
        self.__expect = None
        self.__xfwd = None

        self.channel = None
        self.client = None
        self.agent_forwarding = agent_forwarding
        self.compress = compress
        self.embedded = embedded
        self.no_firsthop = no_firsthop
        self.x11_forwarding = x11_forwarding

        self.stdin = None
        self.stdout = None
        self.stderr = None

        self.__alt_buffer = False
        self.__broadcast = 'off'
        self.__ctrl_fifo = None
        self.__data_fifo = None

        if username:
            # Override username and password from last hop
            data['hops'][-1]['username'] = username
            data['hops'][-1]['password'] = None
            for user in data['users']:
                if user['username'] == username:
                    data['hops'][-1]['password'] = user['password']
                    break

        if broadcast:
            self.do_broadcast('toggle')

    @property
    def afwd(self):
        return self.__afwd is not None

    @property
    def alt_buffer(self):
        return self.__alt_buffer

    @property
    def broadcast(self):
        return self.__broadcast

    @property
    def cfwd(self):
        return self.__cfwd is not None

    @property
    def hostname(self):
        return self.__data['hostname']

    @property
    def mode(self):
        return self.__mode

    @property
    def pfwds(self):
        """ Opened port forwarding tunnels """
        pfwds = [
            [pfwd.index, pfwd.server_host, pfwd.server_port, pfwd.remote_host, pfwd.remote_port]
            for pfwd in self.__pfwds
        ]
        return pfwds

    @property
    def rfwds(self):
        """ Opened reverse port forwarding tunnels """
        rfwds = [
            [rfwd.index, rfwd.server_host, rfwd.server_port, rfwd.remote_host, rfwd.remote_port]
            for rfwd in self.__rfwds
        ]
        return rfwds

    @property
    def title(self):

        flags = []
        if self.__broadcast == 'master':
            flags.append('MASTER')

        elif self.__broadcast == 'slave':
            flags.append('SLAVE')

        if self.__mode:
            flags.append(self.__mode.upper())

        if self.__afwd:
            flags.append('AFWD')

        if self.__cfwd:
            flags.append('CFWD')

        if self.__pfwds:
            flags.append('PFWD')

        if self.__rfwds:
            flags.append('RFWD')

        if self.__xfwd:
            flags.append('XFWD')

        title = 'EeSSH - {0}{1}{2}'.format(
            self.hostname,
            '.{0}'.format(self.__data['customer_uid']) if self.__data['customer_uid'] else '',
            ' [{0}]'.format('|'.join(flags)) if flags else '',
        )

        return title

    @property
    def xfwd(self):
        if self.__xfwd is None:
            return None

        return self.__xfwd.display

    @property
    def username(self):
        transport = self.channel.get_transport()
        return transport.get_username()

    def __process(self):

        buf = b''

        def process_data(data):
            length = len(data)
            if length == 1 and not self.embedded:
                value = ord(data)
                if value == 0:
                    # CTRL+SPACE
                    if self.__broadcast == 'slave':
                        return

                    self.open_eeshell()
                    return

                elif value == 2:
                    # CTRL+B
                    self.do_broadcast('toggle')
                    return

                elif value == 21:
                    # CTRL+U
                    self.do_su()
                    return

            if self.__mode == 'shell':
                # Send data to remote
                if self.channel.send(data) != length:
                    self.__loop = False

                if self.__broadcast == 'master':
                    self.__data_fifo.send(data)

        term_size = None
        while self.__loop:
            # Resize the terminal if needed
            size = self.get_terminal_size()
            if size != term_size:
                self.channel.resize_pty(width=size[1], height=size[0])
                term_size = size

            # Handle input/output
            read_list = [self.channel, self.stdin]

            if self.__broadcast == 'slave':
                read_list.append(self.__ctrl_fifo)
                read_list.append(self.__data_fifo)

            try:
                rlist, wlist, xlist = select.select(read_list, [], [], 0.1)

            except select.error as e:
                if e[0] == errno.EINTR:
                    # May happen when resizing terminal
                    continue
                raise

            if self.channel in rlist:
                data = self.channel.recv(self.READ_SIZE)

                if data == CForwarder.CMD_START:
                    # Copy detection
                    if self.open_cfwd():
                        self.channel.send(CForwarder.CMD_READY + b':' + bytes(str(self.__cfwd.server_port), 'utf-8') + b'\n')

                elif data == CForwarder.CMD_CLOSE:
                    # Copy detection
                    self.close_cfwd()
                    self.channel.send(CForwarder.CMD_CLOSE + b'\n')

                elif self.__mode == 'shell':
                    self.stdout.write(data)
                    self.stdout.flush()

                    if self.__expect:
                        buf += data
                        pattern = self.__expect.get('pattern')
                        if pattern and pattern.search(buf):
                            response = self.__expect.get('response')
                            if response:
                                log.debug('[EXPECT] Buffer matched pattern, sending response')
                                self.channel.send(response)

                            buf = b''
                            self.__expect = None

            if self.stdin in rlist:
                data = self.stdin.read(self.READ_SIZE)
                process_data(data)

            if self.__broadcast == 'slave':
                if self.__data_fifo in rlist:
                    data = self.__data_fifo.recv(self.READ_SIZE)
                    process_data(data)

                if self.__ctrl_fifo in rlist:
                    cmd = self.__ctrl_fifo.recv(self.READ_SIZE).decode('utf-8').rstrip('\n')
                    if cmd == 'broadcast_off':
                        self.do_broadcast('off')
                    elif cmd == 'do_su':
                        self.do_su()
                    else:
                        log.error('[SLAVE] received an invalid command %s', cmd)

            if self.__mode == 'shell' and self.channel.exit_status_ready():
                self.__loop = False

    def copy(self):
        """ Return a copy of this client """

        broadcast = False if self.__broadcast == 'off' else True

        client = EeSSHClient(
            self.__data.copy(),
            agent_forwarding=self.agent_forwarding,
            broadcast=broadcast,
            compress=self.compress,
            embedded=self.embedded,
            x11_forwarding=self.x11_forwarding,
        )

        return client

    def do_broadcast(self, broadcast):
        assert broadcast in ('master', 'off', 'slave', 'toggle'), 'Invalid broadcast mode {0}'.format(broadcast)

        if broadcast == 'toggle':
            if self.__broadcast in ('master', 'slave'):
                broadcast = 'off'
            elif not os.path.exists(self.CTRL_FIFO):
                broadcast = 'master'
            else:
                broadcast = 'slave'

        elif broadcast == self.__broadcast:
            return True

        log.debug('[BROADCAST] changing broadcast mode from %s to %s', self.__broadcast, broadcast)

        if self.__ctrl_fifo:
            if self.__broadcast == 'master':
                # Turning slaves off
                self.__ctrl_fifo.send(b'broadcast_off\n')

            self.__ctrl_fifo.close()
            self.__ctrl_fifo = None

        if self.__data_fifo:
            self.__data_fifo.close()
            self.__data_fifo = None

        self.__broadcast = 'off'

        if broadcast == 'off':
            self.set_title()
            return True

        class_ = MasterFifo if broadcast == 'master' else SlaveFifo

        def create_fifo(filepath, class_=class_):
            try:
                fifo = class_(filepath)
                fifo.start()

            except Exception:
                log.exception('Failed to create fifo file')
                sys.stderr.write('Failed to create fifo file\n')
                return None

            return fifo

        self.__ctrl_fifo = create_fifo(self.CTRL_FIFO)
        if self.__ctrl_fifo is None:
            return False

        self.__data_fifo = create_fifo(self.DATA_FIFO)
        if self.__data_fifo is None:
            self.__ctrl_fifo.close()
            self.__ctrl_fifo = None
            return False

        self.__broadcast = broadcast

        self.set_title()
        return True

    def do_cat(self, path):
        assert self.__sftp, 'Must call EeSSHClient.open_sftp() before calling EeSSHClient.do_cat()'

        try:
            s = self.__sftp.stat(path)
            size = s.st_size
            if size > 512 * KB:
                self.stdout.write(b'File size exceed maximum authorized (512KiB).\n')
                return None

            with self.__sftp.open(path, mode='r', bufsize=1024) as fp:
                content = fp.read()

        except Exception:
            log.exception('Failed to open file')
            return None

        return content

    def do_cd(self, path):
        assert self.__sftp, 'Must call EeSSHClient.open_sftp() before calling EeSSHClient.do_cd()'

        try:
            self.__sftp.chdir(path)

        except Exception:
            log.exception('Failed to change directory')
            return False

        return True

    def do_get(self, remotepath, localpath=None):
        assert self.__sftp, 'Must call EeSSHClient.open_sftp() before calling EeSSHClient.do_get()'

        filename = os.path.basename(remotepath)

        if localpath is None:
            localpath = filename

        elif os.path.isdir(localpath):
            localpath = os.path.join(localpath, filename)

        counts = dict(
            s=time.time(),
            ds=time.time(),
            transfered=0,
        )

        try:
            self.__sftp.get(remotepath, localpath, callback=lambda transfered, total: self.sftp_callback(filename, transfered, total, counts))

        except Exception:
            log.exception('[SFTP] Transfer failed (remotepath=%s, localpath=%s)', remotepath, localpath)
            return False

        finally:
            self.stdout.write(b'\n')

        return True

    def do_ls(self, path):
        assert self.__sftp, 'Must call EeSSHClient.open_sftp() before calling EeSSHClient.do_ls()'

        try:
            data = self.__sftp.listdir_attr(path)

        except Exception:
            log.exception('Failed to list directory %s', path)
            return None

        result = []
        for attr in data:
            attr = attr_data(attr)
            if attr['type'] == 'l':
                attr['link_attr'] = attr_data(self.__sftp.stat(attr['filename']))
                attr['link_attr']['path'] = self.__sftp.readlink(attr['filename'])

            result.append(attr)

        return result

    def do_put(self, localpath, remotepath=None):
        assert self.__sftp, 'Must call EeSSHClient.open_sftp() before calling EeSSHClient.do_put()'

        filename = os.path.basename(localpath)

        if remotepath is None:
            remotepath = filename

        else:
            try:
                s = self.__sftp.stat(remotepath)
                if stat.S_ISDIR(s.st_mode):
                    remotepath = os.path.join(remotepath, filename)

            except Exception:
                pass

        counts = dict(
            s=time.time(),
            ds=time.time(),
            transfered=0,
        )

        try:
            self.__sftp.put(localpath, remotepath, callback=lambda transfered, total: self.sftp_callback(filename, transfered, total, counts))

        except Exception:
            log.exception('[SFTP] Transfer failed (localpath=%s, remotepath=%s)', localpath, remotepath)
            return False

        finally:
            self.stdout.write(b'\n')

        return True

    def do_pwd(self):
        assert self.__sftp, 'Must call EeSSHClient.open_sftp() before calling EeSSHClient.do_pwd()'

        return self.__sftp.getcwd() or '~/'

    def do_su(self):

        # Get parameters from last hop
        data = self.__data['hops'][-1]

        su_method = data.get('su_method')
        if su_method is None:
            self.stdout.write(b'\n\rMissing super user method.\n\r')
            self.channel.send(b'\n')
            return False

        if su_method == 'enable':
            command = 'enable'
            pattern = br'^(Password|Mot de passe)\s*:\s*$'
            response = data.get('su_password')

        elif su_method == 'su':
            command = 'su -'
            pattern = br'^(Password|Mot de passe)\s*:\s*$'
            response = data.get('su_password')

        elif su_method == 'sudo':
            command = 'sudo -i -p \'SUDO PASSWORD:\''
            pattern = br'^SUDO PASSWORD:$'
            response = data.get('su_password') or data.get('password')

        else:
            self.stdout.write(bytes('Invalid root method {0}'.format(su_method), 'utf-8'))
            return False

        self.__expect = dict(
            pattern=re.compile(pattern, re.M),
            response=bytes('{0}\n'.format(response), 'utf-8') if response else None,
        )

        self.channel.send(bytes('{0}\n'.format(command), 'utf-8'))

        if self.__broadcast == 'master':
            self.__ctrl_fifo.send(b'do_su\n')

    def close(self):

        self.do_broadcast('off')

        if self.__afwd:
            # Closing agent forwarding tunnel
            self.__afwd.close()
            self.__afwd = None

        if self.__cfwd:
            # Closing copy tunnel
            self.__cfwd.close()
            self.__cfwd = None

        if self.__xfwd:
            # Closing X11 forwarding tunnel
            self.__xfwd.close()
            self.__xfwd = None

        # Closing port forwarding tunnels
        self.close_pfwd('all')

        # Closing reverse port forwarding tunnels
        self.close_rfwd('all')

        if self.channel:
            self.channel.close()

        # Reset window title
        self.set_title(reset=True)

    def close_cfwd(self):
        """ Close opened copy tunnel with remote """

        if self.__cfwd is None:
            self.stderr.write(b'No tunnel.\n')
            return True

        self.stdout.write(bytes('Closing {0}\n'.format(self.__cfwd.name), 'utf-8'))

        result = self.__cfwd.close()
        self.__cfwd = None

        self.set_title()
        return result

    def close_pfwd(self, index):
        """ Close opened port forwarding tunnel with remote """

        if not self.__pfwds:
            self.stderr.write(b'No tunnel.\n')
            return True

        if index == 'all':
            for pfwd in self.__pfwds:
                self.stdout.write(bytes('Closing port forwarding tunnel from {0}:{1} to {2}:{3}\n'.format(pfwd.server_host, pfwd.server_port, pfwd.remote_host, pfwd.remote_port), 'utf-8'))
                pfwd.close()

            self.__pfwds = []
            self.set_title()
            return True

        found = False
        for pfwd in self.__pfwds:
            if str(pfwd.index) == str(index):
                found = True
                break

        if not found:
            self.stderr.write(b'No such tunnel.\n')
            return True

        self.stdout.write(bytes('Closing port forwarding tunnel from {0}:{1} to {2}:{3}\n'.format(pfwd.server_host, pfwd.server_port, pfwd.remote_host, pfwd.remote_port), 'utf-8'))

        result = pfwd.close()

        self.__pfwds.remove(pfwd)
        self.set_title()
        return result

    def close_rfwd(self, index):
        """ Close opened reverse port forwarding tunnel with remote """

        if not self.__rfwds:
            self.stderr.write(b'No tunnel.\n')
            return True

        if index == 'all':
            for rfwd in self.__rfwds:
                self.stdout.write(bytes('Closing port forwarding tunnel from {0}:{1} to {2}:{3}\n'.format(rfwd.server_host, rfwd.server_port, rfwd.remote_host, rfwd.remote_port), 'utf-8'))
                rfwd.close()

            self.__rfwds = []
            self.set_title()
            return True

        found = False
        for rfwd in self.__rfwds:
            if str(rfwd.index) == str(index):
                found = True
                break

        if not found:
            self.stderr.write(b'No such tunnel.\n')
            return True

        self.stdout.write(bytes('Closing port forwarding tunnel from {0}:{1} to {2}:{3}\n'.format(rfwd.server_host, rfwd.server_port, rfwd.remote_host, rfwd.remote_port), 'utf-8'))

        result = rfwd.close()

        self.__rfwds.remove(rfwd)
        self.set_title()
        return result

    def close_sftp(self):
        """ Close opened SFTP channel with remote """

        if self.__sftp is None:
            return True

        self.__sftp.close()
        self.__sftp = None
        self.set_title()

        return True

    def connect(self):

        if self.client:
            return True

        hops = self.__data['hops'][:]
        if config.get('firsthop', 'enabled') and not self.no_firsthop:
            hops.insert(
                0,
                dict(
                    host=config.get('firsthop', 'hostname'),
                    hostname=config.get('firsthop', 'hostname').split('.')[0],
                    port=config.get('firsthop', 'port'),
                    username=config.get('firsthop', 'username'),
                    password=config.get('firsthop', 'password'),
                )
            )

        sock = None
        for index, data in enumerate(hops):
            if index > 0:
                sock = NCProxy(self.client)
                sock.connect(data['host'], port=data['port'])

            # Enable compression for last hop if requested
            compress = self.compress and index == len(self.__data['hops']) - 1

            self.stdout.write(bytes('[ .... ] Connecting to {username}@{hostname}:{port}'.format(**data), 'utf-8'))

            self.client = SSHClient()
            self.client.set_missing_host_key_policy(AutoAddPolicy())

            try:
                self.client.connect(
                    data['host'],
                    port=data['port'],
                    sock=sock,
                    username=data['username'],
                    password=data.get('password'),
                    compress=compress,
                    timeout=self.CONNECT_TIMEOUT,
                )
            except Exception:
                self.client = None
                self.stderr.write(b'\r[\x1b[31mFAILED\x1b[39m]\n exiting...\n\r')
                log.exception('Failed to connect to %s', data['hostname'])
                return False

            self.stdout.write(b'\r[  \x1b[32mOK\x1b[39m  ]\n\r')

        self.stdout.write(b'\n\n')

        # Open session
        self.channel = self.client.get_transport().open_session()

        # Opening agent forwarding tunnel if needed
        if self.agent_forwarding:
            self.open_afwd()

        # Opening X11 forwarding tunnel if needed
        if self.x11_forwarding:
            self.open_xfwd()

        # Set default path
        data['path'] = '~'

        return True

    def get_terminal_height(self):
        size = self.get_terminal_size()
        return size[0]

    def get_terminal_size(self):
        return struct.unpack('hh', fcntl.ioctl(self.__tty_fd, termios.TIOCGWINSZ, '1234'))

    def get_terminal_width(self):
        size = self.get_terminal_size()
        return size[1]

    def open_afwd(self):
        """ Open a port fowarding tunnel with remote """

        if self.__afwd:
            self.stderr.write(b'A tunnel is already opened.\n\n\r')
            return False

        if self.__shell:
            self.stderr.write(b'A shell is already invoked.\n\n\r')
            return False

        self.__afwd = AgentRequestHandler(self.channel)

        self.set_title()
        return True

    def open_cfwd(self):
        """ Open a copy tunnel with remote """

        if self.__cfwd:
            self.stderr.write(b'A tunnel is already opened.\n\n\r')
            return True

        try:
            cfwd = CForwarder(self.client, dirpath='/tmp')

        except Exception:
            self.stderr.write(bytes('Failed to copy tunnel\n', 'utf-8'))
            return False

        if not cfwd.open():
            self.stderr.write(bytes('Failed to open {0}\n'.format(cfwd.name), 'utf-8'))
            return False

        self.stdout.write(bytes('Openning {0}\n'.format(cfwd.name), 'utf-8'))

        self.__cfwd = cfwd

        self.set_title()
        return True

    def open_eeshell(self):
        """ Open EEShell """

        from eessh.shell import EeCMDShell

        # Enable alternative screen buffer
        self.switch_buffer()

        self.open_sftp()

        cmd = EeCMDShell(self)
        lastcmd = cmd.run()

        self.close_sftp()

        # Disable alternative screen buffer
        self.switch_buffer()

        self.open_shell()

        if lastcmd == 'su':
            self.do_su()

    def open_pfwd(self, server_host, server_port, remote_host, remote_port):
        """ Open a port fowarding tunnel with remote """

        try:
            self.__pfwds_index += 1
            pfwd = PForwarder(self.client, server_host, server_port, remote_host, remote_port, index=self.__pfwds_index)

        except Exception:
            self.stderr.write(bytes('Failed to open port forwarding tunnel from {0}:{1} to {2}:{3}\n'.format(server_host, server_port, remote_host, remote_port), 'utf-8'))
            return False

        if not pfwd.open():
            self.stderr.write(bytes('Failed to open port forwarding tunnel from {0}:{1} to {2}:{3}\n'.format(server_host, server_port, remote_host, remote_port), 'utf-8'))
            return False

        self.stdout.write(bytes('Openning port forwarding tunnel from {0}:{1} to {2}:{3}\n'.format(server_host, server_port, remote_host, remote_port), 'utf-8'))

        self.__pfwds.append(pfwd)

        self.set_title()
        return True

    def open_rfwd(self, server_host, server_port, remote_host, remote_port):
        """ Open a reverse port fowarding tunnel with remote """

        try:
            self.__rfwds_index += 1
            rfwd = RForwarder(self.client, server_host, server_port, remote_host, remote_port, index=self.__rfwds_index)

        except Exception:
            self.stderr.write(bytes('Failed to open port forwarding tunnel from {0}:{1} to {2}:{3}\n'.format(server_host, server_port, remote_host, remote_port), 'utf-8'))
            return False

        if not rfwd.open():
            self.stderr.write(bytes('Failed to open port forwarding tunnel from {0}:{1} to {2}:{3}\n'.format(server_host, server_port, remote_host, remote_port), 'utf-8'))
            return False

        self.stdout.write(bytes('Openning port forwarding tunnel from {0}:{1} to {2}:{3}\n'.format(server_host, server_port, remote_host, remote_port), 'utf-8'))

        self.__rfwds.append(rfwd)

        self.set_title()
        return True

    def open_sftp(self):
        """ Open SFTP channel with remote """

        if self.__sftp:
            self.__mode = 'sftp'
            return True

        if not self.connect():
            return False

        try:
            self.__sftp = self.client.open_sftp()

        except Exception:
            log.exception('Failed to open SFTP')
            return False

        self.__mode = 'sftp'
        self.set_title()
        return True

    def open_shell(self):
        """ Open shell on remote """

        if self.__shell:
            self.__mode = 'shell'
            self.set_title()
            return True

        if not self.connect():
            return False

        # Invoke shell on server
        try:
            size = self.get_terminal_size()

            self.channel.get_pty(term=os.environ.get('TERM', 'xterm'), width=size[1], height=size[0])
            self.channel.invoke_shell()
            self.__shell = True

        except Exception:
            log.exception('Failed to invoke shell on remote')
            return False

        self.__mode = 'shell'
        self.set_title()
        return True

    def open_xfwd(self):
        """ Open a X11 fowarding tunnel with remote """

        if self.__xfwd:
            self.stderr.write(b'A tunnel is already opened.\n\n\r')
            return False

        if self.__shell:
            self.stderr.write(b'A shell is already invoked.\n\n\r')
            return False

        x11_display = os.environ.get('DISPLAY')
        if x11_display is None:
            self.stderr.write(b'Failed to get X11 display, disabling X11 forwarding\n\n\r')
            return False

        try:
            self.__xfwd = XForwarder(self.channel, x11_display)

        except Exception:
            self.stderr.write(bytes('Failed to open X11 forwarding tunnel to {0}\n\n\r'.format(x11_display), 'utf-8'))
            return False

        if not self.__xfwd.open():
            self.stderr.write(bytes('Failed to open X11 forwarding tunnel to {0}\n\n\r'.format(x11_display), 'utf-8'))
            self.__xfwd = None
            return False

        self.stdout.write(bytes('Openning X11 forwarding tunnel to {0}\n\n\r'.format(x11_display), 'utf-8'))

        self.set_title()
        return True

    def prepare_tty(self):
        """
        Disable echo local and set TTY to raw mode so that stdin is ready even if enter is not pressed
        """

        # Disable echo
        mode = termios.tcgetattr(self.__tty_fd)
        mode[tty.LFLAG] &= ~termios.ECHO

        termios.tcsetattr(self.__tty_fd, termios.TCSADRAIN, mode)

        # Set raw mode
        tty.setraw(self.__tty_fd)

    def restore_tty(self):
        """
        Enable echo local and disable raw mode
        """

        termios.tcsetattr(self.__tty_fd, termios.TCSADRAIN, self.__tty_mode)

    def run(self, action='shell', on_ready=None, on_exit=None, tty_fd=None):

        # Reopen standard input/output/error in binary mode
        self.stdin = os.fdopen(sys.stdin.fileno() if tty_fd is None else tty_fd, 'rb', 0)
        self.stdout = os.fdopen(sys.stdout.fileno() if tty_fd is None else tty_fd, 'wb', 0)
        self.stderr = os.fdopen(sys.stderr.fileno() if tty_fd is None else tty_fd, 'wb', 0)

        # Save current TTY mode for later restoration
        self.__tty_fd = self.stdout.fileno()
        self.__tty_mode = termios.tcgetattr(self.__tty_fd)

        self.prepare_tty()

        if not self.connect():
            self.close()
            self.restore_tty()
            if on_exit:
                on_exit(self, False)
            return False

        if action == 'shell':
            self.open_shell()

        elif action == 'eeshell':
            self.open_eeshell()

        else:
            # Invalid action
            self.close()
            self.restore_tty()
            if on_exit:
                on_exit(self, False)
            return False

        self.set_title()

        if on_ready:
            on_ready(self)

        result = True
        try:
            self.__process()
        except Exception:
            log.exception('An exception occured during commmunication with remote')
            result = False

        self.close()
        self.restore_tty()

        if on_exit:
            on_exit(self, result)

        return result

    def set_title(self, reset=False):
        """
        Change window's title
        """

        if self.stdout is None:
            return False

        self.stdout.write(bytes('\033]0;{0}\007'.format('' if reset else self.title), 'utf-8'))

        return True

    def sftp_callback(self, filename, transfered, total, counts):
        """
        SFTP callback to display usefull information during file transfer
        """

        dt = time.time()
        if dt - counts['ds'] >= 1:
            chunk = transfered - counts['transfered']
            speed = chunk // (dt - counts['ds'])
            percent = transfered * 100 // total
            remaining = int((total * (dt - counts['s']) / transfered) - (dt - counts['s']))

            width = self.get_terminal_width()
            txt_r = '{0:3d}% {1} {2}/s {3}'.format(percent, size_hr(transfered), size_hr(speed), time_hr(remaining))
            txt_l = '{0:{1}s}'.format(filename, width - len(txt_r) - 1)

            self.stdout.write(bytes('\r{0} {1}'.format(txt_l, txt_r), 'utf-8'))

            counts['ds'] = dt
            counts['transfered'] = transfered

        if transfered == total:
            speed = total // (dt - counts['s'])

            width = self.get_terminal_width()
            percent = 100
            duration = int(dt - counts['s'])

            txt_r = '{0:3d}% {1} {2}/s {3}'.format(percent, size_hr(transfered), size_hr(speed), time_hr(duration))
            txt_l = '{0:{1}s}'.format(filename, width - len(txt_r) - 1)

            self.stdout.write(bytes('\r{0} {1}'.format(txt_l, txt_r), 'utf-8'))

    def stop(self):
        self.__loop = False

    def switch_buffer(self):
        """ Switch active buffer """

        if not self.__alt_buffer:
            # Enable alternative screen buffer
            self.restore_tty()
            self.stdout.write(b'\x1b[?1049h\x1b[H')

            self.__alt_buffer = True
            return True

        # Disable alternative screen buffer
        self.stdout.write(b'\x1b[?1049l')
        self.prepare_tty()

        self.__alt_buffer = False
        return True


class NCProxy(ClosingContextManager):
    """
    See `paramiko.proxy.ProxyCommand` for original implementation.
    """

    def __init__(self, client):

        self.client = client
        self.host = None
        self.port = 22

        self.stdin = None
        self.stdout = None
        self.stderr = None

        self.timeout = None

    @property
    def closed(self):
        if self.stdin is None:
            return True

        return self.stdin.closed

    @property
    def _closed(self):
        return self.closed

    def close(self):
        if self.stdin is None:
            return True

        self.stdin.close()
        self.stdin = None

        self.stdout.close()
        self.stdout = None

        self.stderr.close()
        self.srderr = None

        return True

    def connect(self, host, port=None):

        self.host = host
        if port:
            self.port = port

        log.info('[PROXY] [%s:%s] Starting NC', self.host, self.port)

        # Start NC on client
        stdin, stdout, stderr = self.client.exec_command('nc -q0 {0} {1}'.format(self.host, self.port))

        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr

    def send(self, content):
        try:
            self.stdin.write(content)

        except IOError:
            log.exception('[PROXY] [%s:%s] IOError while sending data.', self.host, self.port)
            raise

        return len(content)

    def recv(self, size):
        try:
            buf = b''
            while len(buf) < size:
                if self.stdout is None:
                    break

                rlist, wlist, xlist = select.select([self.stdout.channel], [], [], self.timeout)
                if rlist and self.stdout.channel is rlist[0]:
                    buf += self.stdout.read(size - len(buf))

            return buf

        except socket.timeout:
            if buf:
                return buf
            raise

        except IOError:
            log.exception('[PROXY] [%s:%s] IOError while receiving data.', self.host, self.port)
            raise

    def settimeout(self, timeout):
        self.timeout = timeout
