# -*- coding: utf-8 -*-

# EeSSH -- SSH client for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eessh
#
# This file is part of EeSSH.
#
# EeSSH is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EeSSH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from configparser import ConfigParser
from logging.config import fileConfig
import logging
import os
import stat
import keyring


log = logging.getLogger(__name__)


# Constants
APPNAME = 'Easter-Eggs SSH client'
VERSION = '1.0'


class Config:
    appname = APPNAME
    version = VERSION

    def __init__(self):
        self._config = None
        self._defaults = {}
        self._filepath = None

    def get(self, section, option):
        assert self._config is not None, 'Unconfigured options parser'

        methods = {
            bool: self._config.getboolean,
            float: self._config.getfloat,
            int: self._config.getint,
            str: self._config.get,
        }

        type_, default = self.get_default(section, option)
        try:
            value = methods[type_](section, option)
        except Exception:
            value = default

        if section == 'api' and option == 'password' and (value == 'keyring' or value is None):
            username = self.get(section, 'username')
            if not username:
                return None
            value = keyring.get_password('eemyle', username)

        return value

    def get_default(self, section, option):
        assert section in self._defaults, 'Section {0} not registered in defaults configuration'.format(section)
        assert option in self._defaults[section], 'Option {0} not registered in section {1} in defaults configuration'.format(option, section)

        return self._defaults[section][option]

    def load(self, filepath):
        """ Read configuration file """

        self._config = ConfigParser()
        self._filepath = filepath

        # Checking access of configuration file
        if not os.path.isfile(filepath):
            return True

        if not os.access(filepath, os.R_OK):
            return False

        try:
            self._config.read(filepath)
        except Exception:
            self._config = None
            log.exception('Failed to read configuration file %s', filepath)
            return False

        # Logging initialization
        if self._config.has_section('loggers'):
            fileConfig(filepath)

        self._filepath = filepath

        return True

    def save(self):
        """ Save configuration file """

        # Checking access of target directory
        dirpath = os.path.dirname(self._filepath)
        if not os.path.isdir(dirpath) or not os.access(dirpath, os.R_OK | os.W_OK | os.X_OK):
            return False

        if os.path.isfile(self._filepath) and not os.access(self._filepath, os.W_OK):
            return False

        # Make sure config is loaded
        if self._config is None:
            return False

        try:
            with open(self._filepath, 'w') as fp:
                self._config.write(fp)

            # Privacy!
            os.chmod(self._filepath, stat.S_IRUSR | stat.S_IWUSR)
        except Exception:
            log.exception('Failed to write configuration file %s', self._filepath)
            return False

        return True

    def set(self, section, option, value):
        assert self._config is not None, 'Unconfigured options parser'

        if section == 'api' and option == 'password':
            import keyring # late import so that python-keyring is not a mandatory dependency
            username = self.get(section, 'username')
            keyring.set_password('eemyle', username, value)
            value = 'keyring'

        methods = {
            bool: lambda x: str(x).lower(),
            float: lambda x: str(x),
            int: lambda x: str(int(x)),           # Convert float to int
            str: lambda x: x,
        }

        if value == '':
            value = None

        type_, default = self.get_default(section, option)
        if value == default:
            # Remove option from config
            self._config.remove_option(section, option)
        else:
            # Store option to config
            if not self._config.has_section(section):
                self._config.add_section(section)

            value = methods[type_](value)
            self._config.set(section, option, value)

        return True

    def set_defaults(self, section, defaults):
        assert section not in self._defaults, 'Section {0} already register in defaults configuration'.format(section)

        self._defaults[section] = defaults


# Initialize config object
config = Config()
