# -*- coding: utf-8 -*-

# EeSSH -- SSH client for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eessh
#
# This file is part of EeSSH.
#
# EeSSH is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EeSSH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Legacy connector using XML-RPC """

import logging
import threading
import xmlrpc.client

from furl import furl

log = logging.getLogger('eessh.api')

from eessh.cache import xdgcache as region


# Errors
ERROR_UNKNOWN = 10000
ERROR_ACCES_DENIED = 10001
ERROR_TRACEROUTE_NO_SERVER = 10010
ERROR_TRACEROUTE_NO_ROUTE = 10011

marshaller = xmlrpc.client.Marshaller(allow_none=True)


def dump_struct(data):
    """ Simple function to convert a dictionary to a string """

    out = []

    marshaller.dump_struct(data, out.append)

    return b''.join(out)


class LegacyConnectorError(xmlrpc.client.Error):

    def __init__(self):

        xmlrpc.client.Error.__init__(self)
        self.errmsg = 'Unknown error'
        self.errcode = -1

    def __repr__(self):
        return '<LegacyConnectorError: {0} {1}>'.format(self.errmsg, self.errcode)


class LegacyConnector:
    """
    XML-RPC Connector to Eemyle server
    """

    def __init__(self, url, username=None, password=None, debug=False, **kwargs):

        # Build uri
        f = furl(url, username=username, password=password)

        # Build server
        self.__error = None
        self.__server = xmlrpc.client.Server(f.url, verbose=debug)

        self.__lock = threading.Lock()

    def __enter__(self):
        """ Acquire lock """
        self.__lock.acquire()
        return self

    def __exit__(self, type_, value, traceback):
        """ Release lock """
        self.__lock.release()

    @property
    def error(self):
        return self.__error

    @region.cache_on_arguments()
    def get_customers(self):

        result = self.query('eemyle.get_customers_having_server')
        if result is None:
            return None

        customers = []
        for data in result:
            if not data['has_server']:
                continue

            customers.append(dict(
                customer_uid=data['eeid'],
                name=data['name'],
                city=data['city'],
                address=data['street'],
                phone=data['phone'],
                has_server=data['has_server'],
            ))

        return customers

    @region.cache_on_arguments()
    def get_data(self, customer_uid=None, hostname=None, server_id=None):

        if server_id is None:
            servers = self.get_servers(customer_uid)
            if servers is None:
                return None

            server_id = None
            for server in servers:
                if server['hostname'] == hostname:
                    server_id = str(server['id'])
                    break

            if server_id is None:
                return None

        # Get server
        data = self.get_server(server_id)
        if data is None:
            return None

        # Get hops
        hops = self.get_hops(server_id)
        if hops is None:
            return None

        # Get opional IPMI info
        ipmi = self.get_ipmi(server_id)

        data.update(dict(
            hops=hops,
            ipmi=ipmi,
        ))

        return data

    @region.cache_on_arguments()
    def get_hops(self, server_id):

        result = self.query('eemyle.get_server_traceroute_by_server_id', str(server_id))
        if result is None:
            return None

        # Prepare data
        hops = []
        for data in result:
            hop = dict(
                hostname=data.get('server_name'),
                host=data.get('ip'),
                port=int(data.get('ssh_port', '22')),
                username=data.get('admin_name'),
                password=data.get('admin_pass') or None,
            )

            if 'sudo' in data['root_pass'] or data['sudo_root'] == 't':
                hop['su_method'] = 'sudo'
                hop['su_username'] = data['root_name']
                hop['su_password'] = data['admin_pass'] or None

            elif data['root_name'] == 'enable':
                hop['su_method'] = 'enable'
                hop['su_password'] = data['root_pass'] or None

            else:
                hop['su_method'] = 'su'
                hop['su_username'] = data['root_name']
                hop['su_password'] = data['root_pass'] or None

            hops.append(hop)

        return hops

    @region.cache_on_arguments()
    def get_ipmi(self, server_id):

        result = self.query('eemyle.get_server_ipmi_infos_by_server_id', str(server_id))
        if result is None:
            return None

        ipmi = dict(
            host=result['host'] or None,
            port=623,
            username=result['user'] or None,
            password=result['password'] or None,
        )

        ipmi['enabled'] = ipmi['host'] is not None and ipmi['username'] is not None and ipmi['password'] is not None

        return ipmi

    @region.cache_on_arguments()
    def get_server(self, server_id):

        result = self.query('eemyle.get_server_by_server_id', str(server_id))
        if result is None:
            return None

        data = dict(
            id=int(result['id_server']),
            customer_uid=result['id_customer'],
            hostname=result['server_name'].split()[0],
            description=result['usage'],
            comments=result['comments'],
            security=result['security_check'] == 't',
            users=[
                dict(
                    username=user['name'],
                    password=user['pass'] or None,
                )
                for user in result['users']
            ]
        )
        return data

    @region.cache_on_arguments()
    def get_servers(self, customer_uid):

        if not customer_uid:
            return None

        result = self.query('eemyle.get_servers_by_customer_id', customer_uid)
        if result is None:
            return None

        servers = []
        for data in result:
            if data['dead'] == 't':
                continue

            servers.append(dict(
                id=int(data['id_server']),
                hostname=data['server_name'].split()[0],
                hoster=None,
                os_name=data['os_name'],
                description=data['usage'],
                security=data['security_check'] == 't',
                usage=', '.join(sorted(data['usages'].values()) if data['usages'] else []),
                usernames=sorted(data['users'].values()) if data['users'] else [],
            ))

        return servers

    @region.cache_on_arguments()
    def get_tickets(self, customer_uid=None, hostname=None, server_id=None):

        if server_id is None:
            servers = self.get_servers(customer_uid)
            if servers is None:
                return None

            server_id = None
            for server in servers:
                if server['hostname'] == hostname:
                    server_id = str(server['id'])
                    break

            if server_id is None:
                return None

        result = self.query('eemyle.get_tickets_by_server_id', str(server_id))
        if result is None:
            return None

        tickets = []
        for data in result:
            tickets.append(data)

        return tickets

    def query(self, function_name, *args):
        """ Sends a request to the server """

        self.__error = None

        # Encoding the arguments
        enc_args = []
        for arg in args:

            if isinstance(arg, dict):
                enc_arg = dump_struct(arg)
            elif isinstance(arg, int):
                enc_arg = str(arg)
            else:
                try:
                    enc_arg = xmlrpc.client.Binary(bytes(arg, 'utf-8'))
                except Exception as e:
                    log.exception('Failed to encode argument {} of type {} for function {}({})'.format(str(arg), type(arg), function_name, ",".join([str(i) for i in args])))
                    enc_arg = str(arg)

            enc_args.append(enc_arg)

        # Calling the function
        try:
            function = getattr(self.__server, function_name)
            result = function(*enc_args)

        except xmlrpc.client.Error as e:
            # Xmlrpclib error
            log.exception('Failed to call xmlrpc function {}({})'.format(function_name, ",".join([str(i) for i in args])))

            self.__error = e
            result = None

        except Exception:
            log.exception('Failed to call xmlrpc function {}({})'.format(function_name, ",".join([str(i) for i in args])))

            # Unknown error
            self.__error = LegacyConnectorError()
            result = None

        return result
