# -*- coding: utf-8 -*-

# EeSSH -- SSH client for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eessh
#
# This file is part of EeSSH.
#
# EeSSH is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EeSSH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Shell """

from functools import cmp_to_key
import logging
import os
import os.path
import shlex
import stat

from cmd2 import Cmd

from eessh.utils import attr_repr
from eessh.utils import attr_sort
from eessh.utils import get_host_port


log = logging.getLogger('eessh.shell')

COMMAND_PROMPT = '(\x1b[34meessh\x1b[39m) {username}@{hostname}:{path} > '
DIRPATH = os.path.join(os.path.expanduser('~'), '.config', 'eemyle') # XXX hard-coded path (and repeated elsewhere)


class EeCMDShell(Cmd):
    allow_cli_args = False

    def __init__(self, client):

        self.path = '~'

        self.client = client
        self.prompt = COMMAND_PROMPT.format(
            username=client.username,
            hostname=client.hostname,
            path=self.path,
        )

        if not os.path.isdir(DIRPATH):
            try:
                os.makedirs(DIRPATH)
            except OSError:
                pass

        # Calling inherited
        Cmd.__init__(self, persistent_history_file=os.path.join(DIRPATH, 'history'))

    def complete_remote(self, path, text, dir_only=False):

        # Extract directory from path
        path = os.path.dirname(path) or '.'

        log.debug('[COMPLETE] type=remote, path=%s', path)

        # List directory
        data = self.client.do_ls(path)
        if data is None:
            return []

        # Filter result
        result = []
        for attr in data:
            filename = attr.filename
            if text and not filename.startswith(text):
                continue

            isdir = stat.S_ISDIR(attr.st_mode)
            if dir_only and not isdir:
                continue

            if isdir:
                filename += '/'

            result.append(filename)

        if len(result) == 1 and not isdir and not dir_only:
            result[0] += ' '

        return result

    def complete_local(self, path, text, dir_only=False):

        # Extract directory from path
        path = os.path.dirname(path) or '.'

        log.debug('[COMPLETE] type=local, path=%s', path)

        # List directory and filter result
        result = []
        for filename in os.listdir(path):
            if text and not filename.startswith(text):
                continue

            fullpath = os.path.join(path, filename)
            isdir = os.path.isdir(fullpath)
            if dir_only and not isdir:
                continue

            if isdir:
                filename += os.sep

            result.append(filename)

        if len(result) == 1 and not isdir and not dir_only:
            result[0] += ' '

        return result

    def complete_cat(self, text, line, begidx, endidx):

        args = shlex.split(line)

        # Get argument to autocomplete
        arg = ''
        idx = 0
        for index, arg in enumerate(args):
            idx += len(arg)
            if idx > begidx:
                break

        return self.complete_remote(arg, text)

    def complete_cd(self, text, line, begidx, endidx):

        args = shlex.split(line)

        # Get argument to autocomplete
        arg = ''
        idx = 0
        for index, arg in enumerate(args):
            idx += len(arg)
            if idx > begidx:
                break

        return self.complete_remote(arg, text, dir_only=True)

    def complete_get(self, text, line, begidx, endidx):

        args = shlex.split(line)

        # Get argument to autocomplete
        arg = ''
        idx = 0
        index = 0
        for index, arg in enumerate(args):
            idx += len(arg)
            if idx > begidx:
                break

        if index == 0 or (index == 1 and begidx <= idx + 1):
            return self.complete_remote(arg, text)

        if index == 1 and idx < begidx:
            arg = ''

        return self.complete_local(arg, text)

    def complete_ls(self, text, line, begidx, endidx):
        return self.complete_cd(text, line, begidx, endidx)

    def complete_pfwd(self, text, line, begidx, endidx):

        args = shlex.split(line)

        if len(args) == 1 or (len(args) == 2 and text):
            options = [
                option
                for option in ['close', 'list', 'open']
                if not text or option.startswith(text)
            ]

            return options

        if (len(args) == 2 and args[1] == 'close') or (len(args) == 3 and text):
            options = [
                option
                for option in ['all'] + [str(pfwd[0]) for pfwd in self.client.pfwds]
                if not text or option.startswith(text)
            ]

            return options

        return []

    def complete_put(self, text, line, begidx, endidx):

        args = shlex.split(line)

        # Get argument to autocomplete
        arg = ''
        idx = 0
        index = 0
        for index, arg in enumerate(args):
            idx += len(arg)
            if idx > begidx:
                break

        if index == 0 or (index == 1 and begidx <= idx + 1):
            return self.complete_local(arg, text)

        if index == 1 and idx < begidx:
            arg = ''

        return self.complete_remote(arg, text)

    def complete_rfwd(self, text, line, begidx, endidx):

        args = shlex.split(line)

        if len(args) == 1 or (len(args) == 2 and text):
            options = [
                option
                for option in ['close', 'list', 'open']
                if not text or option.startswith(text)
            ]

            return options

        if (len(args) == 2 and args[1] == 'close') or (len(args) == 3 and text):
            options = [
                option
                for option in ['all'] + [str(rfwd[0]) for rfwd in self.client.rfwds]
                if not text or option.startswith(text)
            ]

            return options

        return []

    def do_su(self, arg):
        """ Become super user """

        # Command will be performed by client itself
        self.lastcmd = 'su'
        return True

    def do_exit(self, arg):
        """ Exit from command mode """

        return True

    def do_cat(self, arg):
        """cat <remote_file> - copy file content to stdout (SFTP)"""

        args = shlex.split(arg)

        if not len(args) == 1:
            self.stdout.write('Misssing argument.\n')
            return

        path = args[0]

        content = self.client.do_cat(path)
        if content is None:
            self.stdout.write('Failed to open file.\n')
            return

        self.stdout.write('{0}\n'.format(content))

    def do_cd(self, arg):
        """ cd <path> - change current directory on remote (SFTP) """

        args = shlex.split(arg)

        path = args[0] if args else None
        if path == '~':
            path = None

        if not self.client.do_cd(path):
            self.stdout.write('Failed to change directory.\n')
            return

        self.path = '~' if path is None else os.path.normpath(os.path.join(self.path, path))

        # Update prompt
        self.prompt = COMMAND_PROMPT.format(
            username=self.client.username,
            hostname=self.client.hostname,
            path=self.path,
        )

    def do_get(self, arg):
        """get <remote_file> <local_file> - get file from remote (SFTP)"""

        args = shlex.split(arg)

        if not args:
            self.stdout.write('Missing argument.\n')
            return

        remotepath = args[0]
        localpath = args[1] if len(args) > 1 else None

        if not self.client.do_get(remotepath, localpath):
            self.stdout.write('Failed to transfer file.\n')

    def do_ls(self, arg):
        """ List files from current directory on remote (SFTP) """

        args = shlex.split(arg)
        path = args[0] if args else '.'

        data = self.client.do_ls(path)
        if data is None:
            self.stdout.write('Failed to list directory.\n')
            return

        for attr in sorted(data, key=cmp_to_key(attr_sort)):
            self.stdout.write('{0}\n'.format(attr_repr(attr)))

    def do_put(self, arg):
        """put <local_file> <remote_file> - put file to remote (SFTP)"""

        args = shlex.split(arg)

        if not args:
            self.stdout.write('Missing argument.\n')
            return

        localpath = args[0]
        remotepath = args[1] if len(args) > 1 else None

        if not self.client.do_put(localpath, remotepath):
            self.stdout.write('Failed to transfer file.\n')

    def do_pwd(self, arg):
        """ Display current directory on remote (SFTP) """

        path = self.client.do_pwd()

        self.stdout.write('{0}\n\n'.format(path))

    def do_pfwd(self, arg):
        """
        Open or close a port forwarding tunnel with remote

        Usage:
            pfwd [list]
            pfwd open [server_host:]server_port [remote_host:]remote_port
            pfwd close index

        server_host: optional server host to listen on, default is to listen on localhost
        server_port: server port to listen on
        remote_host: optional remote host to connect to, default is to connect to localhost
        remote_port: remote port to connect to

        Without arguments, `pfwd' list current opened tunnel(s).
        """

        args = shlex.split(arg)

        if not args or args[0] == 'list':
            pfwds = self.client.pfwds
            if not pfwds:
                self.stdout.write('No tunnel.\n')
                return

            for index, server_host, server_port, remote_host, remote_port in pfwds:
                self.stdout.write('[{0:3d}] from {1}:{2} to {3}:{4}\n'.format(index, server_host, server_port, remote_host, remote_port))

        elif args[0] == 'close':
            if len(args) < 2:
                self.stdout.write('Missing argument.\n')
            else:
                self.client.close_pfwd(args[1])

        elif args[0] == 'open':
            if len(args) < 3:
                self.stdout.write('Missing argument.\n')
            else:
                try:
                    server_host, server_port = get_host_port(args[1])
                    remote_host, remote_port = get_host_port(args[2])

                except ValueError:
                    self.stdout.write('Invalid argument.\n')

                else:
                    self.client.open_pfwd(server_host, server_port, remote_host, remote_port)
        else:
            self.stdout.write('Invalid argument.\n')

    def do_rfwd(self, arg):
        """
        Open or close a reverse port forwarding tunnel with remote

        Usage:
            rfwd [list]
            rfwd open [server_host:]server_port [remote_host:]remote_port
            rfwd close index

        server_host: optional server host to listen on, default is to lisen on localhost
        server_port: server port to listen on
        remote_host: optional remote host to connect to, default is to connect to localhost
        remote_port: remote port to connect to

        Without arguments, `rfwd' list current opened tunnel(s).
        """

        args = shlex.split(arg)

        if not args or args[0] == 'list':
            rfwds = self.client.rfwds
            if not rfwds:
                self.stdout.write('No tunnel.\n')
                return

            for index, server_host, server_port, remote_host, remote_port in rfwds:
                self.stdout.write('[{0:3d}] from {1}:{2} to {3}:{4}\n'.format(index, server_host, server_port, remote_host, remote_port))

        elif args[0] == 'close':
            if len(args) < 2:
                self.stdout.write('Missing argument.\n')
            else:
                self.client.close_rfwd(args[1])

        elif args[0] == 'open':
            if len(args) < 3:
                self.stdout.write('Missing argument.\n')
            else:
                try:
                    server_host, server_port = get_host_port(args[1])
                    remote_host, remote_port = get_host_port(args[2])

                except ValueError:
                    self.stdout.write('Invalid argument.\n')

                else:
                    self.client.open_rfwd(server_host, server_port, remote_host, remote_port)
        else:
            self.stdout.write('Invalid argument.\n')

    def emptyline(self):
        # Don't repeat last command
        return ''

    def get_names(self):
        # Prevent from having «EOF» in commands list
        names = [name for name in Cmd.get_names(self) if name not in ('do_eof', 'do_eos', 'do__relative_load')]
        return names

    def run(self):

        intro = 'Welcome to the EeSSH shell. Type help or ? to list commands.\n'

        loop = True
        while loop:
            try:
                self.cmdloop(intro=intro)

            except KeyboardInterrupt:
                intro = '\n'

            else:
                loop = False

        return self.lastcmd
