# -*- coding: utf-8 -*-

# EeSSH -- SSH client for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eessh
#
# This file is part of EeSSH.
#
# EeSSH is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EeSSH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import datetime
import errno
import logging
import os
import socket
import stat
import threading


log = logging.getLogger(__name__)


KB = 1024
MB = KB * KB
GB = MB * KB
TB = GB * KB
PB = TB * KB


class MasterFifo(threading.Thread):
    type = 'master'

    def __init__(self, filepath):

        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.bind(filepath)

        os.chmod(filepath, 0o600)

        self.__clients = []
        self.__filepath = filepath
        self.__loop = False

        # Calling inherited
        super(MasterFifo, self).__init__()

        # Non-blocking mode
        self.daemon = True

    def close(self):

        self.__loop = False

        self.sock.close()

        os.remove(self.__filepath)

    stop = close

    def run(self):

        self.__loop = True

        self.sock.listen(1)
        while self.__loop:
            client, info = self.sock.accept()
            self.__clients.append(client)

    def send(self, data):
        for client in self.__clients[:]:
            try:
                client.send(data)
            except IOError as e:
                if e.errno == errno.EPIPE:
                    log.info('[BROADCAST] Client closed connection')
                    self.__clients.remove(client)
                    continue

                log.exception('[BROADCAST] failed to send data to client')


class SlaveFifo(socket.socket):
    type = 'slave'

    def __init__(self, filepath):

        # Calling inherited
        super(SlaveFifo, self).__init__(socket.AF_UNIX, socket.SOCK_STREAM)

        self.__filepath = filepath

    def read(self, length):
        return self.recv(length)

    def start(self):
        self.connect(self.__filepath)

    def stop(self):
        return self.close()

    def write(self, data):
        return self.send(data)


def attr_data(attr):
    def rwx(n, suid, sticky=False):
        if suid:
            suid = 2
        out = "-r"[n >> 2] + "-w"[(n >> 1) & 1]
        if sticky:
            out += "-xTt"[suid + (n & 1)]
        else:
            out += "-xSs"[suid + (n & 1)]
        return out

    data = dict()
    if attr.st_mode is not None:
        kind = stat.S_IFMT(attr.st_mode)
        if kind == stat.S_IFIFO:
            data['type'] = 'p'
        elif kind == stat.S_IFCHR:
            data['type'] = 'c'
        elif kind == stat.S_IFDIR:
            data['type'] = 'd'
        elif kind == stat.S_IFBLK:
            data['type'] = 'b'
        elif kind == stat.S_IFREG:
            data['type'] = '-'
        elif kind == stat.S_IFLNK:
            data['type'] = 'l'
        elif kind == stat.S_IFSOCK:
            data['type'] = 's'
        else:
            data['type'] = '?'

        data['mode'] = attr.st_mode

        data['owner_permissions'] = rwx((attr.st_mode & 0o700) >> 6, attr.st_mode & stat.S_ISUID)
        data['group_permissions'] = rwx((attr.st_mode & 0o70) >> 3, attr.st_mode & stat.S_ISGID)
        data['other_permissions'] = rwx(attr.st_mode & 7, attr.st_mode & stat.S_ISVTX, True)

    else:
        data['type'] = '?'
        data['permissions'] = None
        data['owner_permissions'] = '---'
        data['group_permissions'] = '---'
        data['other_permissions'] = '---'

    # compute display date
    if (attr.st_mtime is None) or (attr.st_mtime == 0xffffffff):
        # shouldn't really happen
        data['mtime'] = None
    else:
        data['mtime'] = datetime.datetime.fromtimestamp(attr.st_mtime)

    data['filename'] = getattr(attr, 'filename', '?')

    # not all servers support uid/gid
    data['uid'] = attr.st_uid or 0
    data['gid'] = attr.st_gid or 0

    data['size'] = attr.st_size or 0

    return data


def attr_repr(data):
    """
    Slightly modified version of SFTPAttributes.__str__() with unicode support and human readable size
    """

    data = data.copy()

    data['size_hr'] = size_hr(data['size'])
    data['mtime_hr'] = date_hr(data['mtime'])
    data['link_info'] = ' -> {0}'.format(data['link_attr']['path']) if data['type'] == 'l' else ''

    return '{type}{owner_permissions}{group_permissions}{other_permissions} 1 {uid:<8d} {gid:<8d} {size_hr:>12s} {mtime_hr:>15s} {filename}{link_info}'.format(**data)


def attr_sort(a1, a2):
    """ Sort function for SFTP attributes objects """

    # Directory first
    if a1['type'] == 'd':
        if not a2['type'] == 'd':
            return -1

    elif a2['type'] == 'd':
        return 1

    f1 = a1['filename'].lower()
    f2 = a2['filename'].lower()

    return (f1 > f2) - (f1 < f2)


def date_hr(date):

    today = datetime.datetime.today()
    delta = today - date

    if abs(delta.days) > 182:
        # 6 months
        format_ = '%-2d %5b %-5Y'
    else:
        format_ = '%-2d %5b %H:%M'

    return date.strftime(format_)


def get_host_port(spec):

    if ':' in spec:
        host, port = spec.split(':')

    else:
        host = '127.0.0.1'
        port = spec

    port = int(port)
    return (host, port)


def size_hr(size):

    if not size:
        return '0'

    if size >= PB:
        return '{0:7.2f}P'.format(size / PB)

    if size >= TB:
        return '{0:7.2f}T'.format(size / TB)

    if size >= GB:
        return '{0:7.2f}G'.format(size / GB)

    if size >= MB:
        return '{0:7.2f}M'.format(size / MB)

    if size >= KB:
        return '{0:7.2f}K'.format(size / KB)

    return '{0}B'.format(size)


def time_hr(duration):

    if duration > 86400:
        # > 1d
        return 'STALLED'

    hr = []

    # hours
    if duration > 3600:
        hr.append('{0:02d}'.format(duration // 3600))
        duration = duration % 3600

    # minutes
    hr.append('{0:02d}'.format(duration // 60))
    duration = duration % 60

    # seconds
    hr.append('{0:02d}'.format(duration))

    return ':'.join(hr)
